pub mod vector;

pub mod simd;
pub mod memory;
pub mod profiling;

//pub use vector::{SimilarityMetric, compute_similarity, cosine_similarity, dot_product, euclidean_distance};
//pub use simd::{create_simd_engine, compute_similarity_simd, SimdEngine};
//pub use memory::{ThreadLocalBuffer, BufferPool, MemoryTracker};
//pub use profiling::{PerformanceMetrics, QueryProfiler};